/**
 * channels module
 */
tma.registerModule( channelsort );
channelsort.prototype = new tmaModule();
function channelsort() {

	/**
	 * Set to true, if ChannelListSaver.tap runs in TSR mode
	 * In TSR mode, CLS is not able to signal im-/export success
	 * currently, CLS v1.0 is configured as non-TSR-TAP
	 */
	var cls_tsr_mode = false;

	/**
	 * minimum required CLS version
	 */
	var Min_CLS_Version = '1.0';

	/**
	 * minimum required SmartEPG version
	 * since this 7.3, SE is able to recognise changed channel list through a state file
	 */
	var Min_SE_Version = '7.3';

	/**
	 * IMPORTANT: the modules name; equal to directory and script file name
	 */
	this.name = 'channelsort';

	this.pageGroup = 'channels';

	this.dependencies = {

			css: [
			      "channelsort.css"
			      ],

			js: [
					"../../js/vendor/shortcut.js",
					"FileSaver.min.js"
			     ],

			templates: [
			            "content.html",
						"head.html",
			            "update.html"
			            ]
	};

	var self = this;

	var CLSdirectory = "/mnt/hd/ProgramFiles/Settings/ChannelListSaver";
	var CLStapid = 705298434;
	var CLSfilename = ''; // filled by checkCLSversion

	var CRLF = "\r\n";

	var cls = {
			text: null,
			FileVersion: 0,
			SystemType: 0,
			UTF8System: 0,
			NrSatellites: 0,
			NrTransponders: 0,
			NrFavsPerGroup: 0,
			slices : {
				beforeTV: {start:0,end:0},
				beforeRadio: {start:0,end:0},
				beforeFav: {start:0,end:0},
			},
	};

	var undoMgr = JK.model({
			stack: [],
			undoPtr: 0,
			redoPtr: 0,
	});

	/**
	 * class: undo element
	 */
	var Undo = (function() {

		function Undo( undo ) {

			// call super class constructor
			Undo._super.constructor.apply(this);

			// initialise

			this.init( undo );

		}
		JK.inherit(Undo, JKobject);

		Undo.define({
			action			: '',
			position		: 0,
			channel			: null,
			list			: null,
		});

		Undo.DELETE = 'delete';
		Undo.INSERT = 'insert';

		return Undo;

	})();

	/**
	 * class: sortable channel object
	 *
	 * @param {ClsChannel}	clsChannel
	 */
	var ClsChannel = (function() {

		function ClsChannel( clsChannel ) {

			// call super class constructor
			ClsChannel._super.constructor.apply(this);

			// initialise

			this.init( clsChannel );

		}
		JK.inherit(ClsChannel, JKobject);

		ClsChannel.define({
			clsLine			: null,
			number			: 0,
			originalNr		: 0,
			selected		: false,
			filtered		: false,
			channel			: null,
		});

		return ClsChannel;

	})();

	/**
	 * class: sortable channel list
	 *
	 * @param {ClsChannel}	clsChannel
	 */
	var ClsChannels = (function() {

		function ClsChannels() {

			// call super class constructor
			ClsChannels._super.constructor.apply(this);

			// initialise
			this.favGroup = '';

			// internal representation of fav group
			this.chtype = 0;

			// number of selected channels
			this.selected = 0;

			// recently selected .cls-channel in channel list
			this.recentSelected = null;

			// regular expression object or false, if unfiltered
			this.textFilter = false;

			// 0: not allowed; 1: display FTA channels; 2: display CAS channels; 3: display all
			this.casFilter = 3;

			this.channels = [];

			this.init();

		}
		JK.inherit(ClsChannels, JKobject);

		ClsChannels.prototype.reset = function() {

			this.channels.length = 0;
			this.recentSelected = null;
			this.textFilter = false;

		};

		/**
		 * get a cls-channel object through channelid
		 *
		 * @param {String}	channelid (logoid)
		 * @returns {ClsChannel} or null, if not found
		 */
		ClsChannels.prototype.channel = function( channelid ) {

			return this.channels.find( function( clsChannel ) {
				return clsChannel.channel.logoid == channelid;
			});

		};

		/**
		 * toggle the selected state of a cls-channel object
		 *
		 * @param {ClsChannel}	clsChannel
		 */
		ClsChannels.prototype.toggleSelected = function( clsChannel ) {

			if ( clsChannel ) {
				if ( clsChannel.selected ) {
					this.selected--;
					clsChannel.selected = false;
				} else {
					this.selected++;
					clsChannel.selected = true;
				}
			}

		};

		/**
		 * set the selected state of all visible cls-channels
		 */
		ClsChannels.prototype.selectAll = function() {

			this.selected = 0;
			for ( var i = 0, len = this.channels.length; i < len; i++ ) {
				var clsChannel = this.channels[i];
				if ( clsChannel.channel.chtype === this.chtype ) {
					clsChannel.selected = true;
					this.selected++;
				}
			}

		};

		/**
		 * remove the selected state of all cls-channels
		 */
		ClsChannels.prototype.deselectAll = function() {

			for ( var i = 0, len = this.channels.length; i < len; i++ ) {
				this.channels[i].selected = false;
			}
			this.selected = 0;

		};

		/**
		 * get a list of all selected cls-channels
		 *
		 * @returns {Array}
		 */
		ClsChannels.prototype.getSelected = function() {
			var result = [];

			for ( var i = 0, len = this.channels.length; i < len; i++ ) {
				var clsChannel = this.channels[i];
				if ( clsChannel.selected ) {
					result.push( clsChannel );
				}
			}

			return result;

		};

		/**
		 * re-number channels
		 *
		 * @param {Number}	index	start index
		 */
		ClsChannels.prototype.renumber = function( index ) {

			var startIndex = index || 0;
			var nr = 1;
			for( var i = startIndex, len = this.channels.length; i < len; i++ ) {
				if ( this.channels[i].channel.chtype === this.chtype ) {
					// only renumber current channel type
					this.channels[i].number = nr++;
				}
			}

		};

		/**
		 * toggle the selected state of a cls-channel object
		 *
		 * @param {ClsChannel}	clsChannel
		 * @returns {Undo}	or false
		 */
		ClsChannels.prototype.deleteChannel = function( clsChannel ) {

			var index = this.channels.findIndex( function( clsCh ) {
				return clsCh === clsChannel;
			});

			if ( index > -1 ) {
				// adapt selected
				if ( clsChannel.selected ) {
					this.selected--;
				}

				// remove channel from list
				this.channels.splice(index, 1);

				// create undo object
				return new Undo({
						action: Undo.DELETE,
						position: index,
						channel: clsChannel,
						list: this,
				});
			}
			return false;
		};

		/**
		 * add a cls-channel object to channels list
		 *
		 * @param {ClsChannel}	clsChannel
		 * @param {Number}		position	optional: insert before this index
		 */
		ClsChannels.prototype.insert = function( clsChannel, position ) {

			if ( clsChannel ) {

				// adapt selected
				if ( clsChannel.selected ) {
					this.selected++;
				}

				this.filterChannel( clsChannel );

				// insert channel
				if ( typeof position == 'undefined') {
					this.channels.push( clsChannel );
				} else {
					this.channels.splice( position, 0, clsChannel );
				}

				// create undo object
				return new Undo({
						action: Undo.INSERT,
						position: position,
						channel: clsChannel,
						list: this,
				});

			}

		};

		ClsChannels.prototype.filterChannel = function( clsChannel ) {

			var channel = clsChannel.channel;
			if ( channel.chtype === this.chtype ) {
				// only filter current channel type
				clsChannel.filtered = !(
						(
								!this.textFilter ||
								!!channel.servicename.match( this.textFilter ) ||
								!!channel.providername.match( this.textFilter )
						) &&
						(((channel.flagcas + 1) & this.casFilter) > 0)

				);
				// adapt selected
				if ( clsChannel.filtered && clsChannel.selected ) {
					clsChannel.selected = false;
					this.selected--;
				}
			}

		};

		ClsChannels.prototype.filter = function( filterText, filterCas ) {

			if ( typeof filterText == 'string' ) {
				if ( filterText ) {
					this.textFilter = new RegExp( filterText, 'i' );
				} else {
					this.textFilter = false;
				}
			}
			if ( typeof filterCas == 'number' ) {
				this.casFilter = filterCas;
			}

			for( var i = 0, len = this.channels.length; i < len; i++ ) {
				this.filterChannel( this.channels[i] );
			}


		};

		ClsChannels.mixin({
			favGroupChange: function() {
				"watch: {{favGroup}}";
				this.chtype = (this.favGroup == '__tv') ? 0 : 1;
				this.deselectAll();
			},
		});

		return ClsChannels;

	})();

	this.channels = new ClsChannels();
	this.clipboard = new ClsChannels();

	this.channels.on('change', 'favGroup', function(e, val) {
		// clipboard favGroup follows channel list's favGroup
		self.clipboard.favGroup = val;
	} );

	var defaultUpdateStatus = {
		saveTextfile:	'none',
		refreshTimers:	'none',
		seWasRunning:	false,
		stopSE:			'none',
		startCLS:		'none',
		startSE:		'none',
		createTimers:	'none',
//		syncSE:			'none',
		prevTimersCnt:	0,
		postTimersCnt:  0,
		clsResult:		true,
	};

	var updateStatus = JK.model( defaultUpdateStatus );
	updateStatus.extend({
		recreated_timers_fmt: function() {
			"watch: {{postTimersCnt}},{{prevTimersCnt}}";
			return self.i18.update.recreated_timers.format(this.postTimersCnt, this.prevTimersCnt);
		}
	});

	var $contentTmpl, $stickyTmpl;

	/**
	 * initialize module
	 */
	this.setup = function() {

		// register tmaFeatures
		tma.fn.addFeature( this, 'appendMenuTab' );
		tma.fn.addFeature( this, 'openChannelInfo' );

		return this.loadLanguage().done(function() {

			// extend TmaChannel
			if (!TmaChannel.prototype.typeStr) {
				TmaChannel.prototype.typeStr = function() {
					switch (this.chtype) {
						case  1: return self.radio		|| 'Radio';
						case -1: return self.analogue	|| 'Analogue';
						default: return self.tv			|| 'TV';
					}
				};
			}

		});

	};

	/**
	 * append tab to menu
	 *
	 * @param {tmaFeature}	f
	 */
	this.appendMenuTab = function( f ) {

		tma.call('addPage', ['MainMenu', this]);

	};

	this.addShortcuts = function() {

		shortcut.add("Ctrl+x"		, function() { $('.cls-but-cut').click(); });
		shortcut.add("Ctrl+v"		, function() { $('.cls-but-paste-before').click(); });
//		shortcut.add("Down"			, function() {
//			var $selected = $('.cls-selected').first();
//			if ( $selected.length ) {
//				// move selection down
//				$selected.next('.cls-channel-ctx').first().click();
//			} else {
//				// select first channel
//				$('.cls-channel-ctx').first().click();
//			}
//		});
//		shortcut.add("Up"			, function() {
//			var $selected = $('.cls-selected').first();
//			if ( $selected.length ) {
//				// move selection up
//				$selected.prev('.cls-channel-ctx').first().click();
//			}
//		});

	};

	this.removeShortcuts = function() {

		shortcut.remove("Ctrl+x");
		shortcut.remove("Ctrl+v");
//		shortcut.remove("Down");
//		shortcut.remove("Up");

	};

	/**
	 * this page gets removed => remove databinding templates, etc.
	 */
	this.cleanup = function() {

		JK.removeTemplate($contentTmpl);
		JK.removeTemplate($stickyTmpl);

		//this.channels.favGroup = '';

		$(window).off('resize.channelsort');

		// remove shortcuts
		this.removeShortcuts();

		$('#timestamp').show();

	};

	var CLS_EXPORT	= 0;
	var CLS_IMPORT	= 1;
	var CLS_FMT_BIN	= 0;
	var CLS_FMT_TXT	= 2;
	var CLS_FMT_SYS	= 4;
	var CLS_VERBOSE	= 0;
	var CLS_SILENT	= 8;
	var CLS_SUCCESS	= 128;
	// errors
	var CLS_FAILED  = -1;
	var CLS_TIMEOUT = -2;
	var CLS_ERROR	= -3;

	/**
	 * create TAP start parameter data block
	 *
	 * @param {String}	filename
	 * @param {Number}	doImport	1: import
	 * 								0: export
	 * @param {Number}	format		0: Binary
	 * 								2: Text
	 * 								4: System
	 * @param {Number}	silent		0: show confirmation dialog
	 * 								1: skip confirmation dialog
	 * @returns Uint8Array
	 */
	this.getClsTapParams = function (filename, doImport, format, silent) {

		/* ChannelListSaver parameter struct:
		typedef struct {
		  byte ImExport:1;       // 1 for Import, 0 for Export  (bit 0)
		  byte Format:2;         // 0 - Binary, 1 - Text, 2 - System  (bit 1,2)
		  byte SilentMode:1;     // 0 - show confirmation dialog, 1 - skip confirmation dialog  (bit 3)
		  byte Reserved:4;
		  char FileName[MAX_FILE_NAME_SIZE];
		};
		*/

		var ui8 = Uint8Array.createFromString( 'x' + filename + 'X' );

		// replace first byte with command bits
		ui8[0] = doImport + format + silent;

		// replace last byte (X) with c string delimiter
		ui8[ui8.length - 1] = 0;

		return ui8;

	};

	this.decodeClsTapParams = function( base64params ) {

		var binParams = Uint8Array.base64decode(base64params);

		var result = {
				import:		binParams[0] & 0x01,
				format: 	binParams[0] & 0x06,
				silentMode:	binParams[0] & 0x08,
				result:		binParams[0] & 0x80,
				filename:   ''
		};

		for( var i = 1, len = binParams.length; (i < len) && (binParams[i] !== 0); i++) {
			result.filename += String.fromCharCode( binParams[i] );
		}

		return result;

	};

	/**
	 * start ChannelListSaver
	 *
	 * @param {String}	filename	file to im/export
	 * @param {Number}	doImport	1: import
	 * 								0: export
	 * @param {Number}	format		0: Binary
	 * 								2: Text
	 * 								4: System
	 * @param {Number}	silent		0: show confirmation dialog
	 * 								1: skip confirmation dialog
	 * @returns {jQueryDeferred}
	 */
	this.startCLS = function(filename, doImport, format, silent) {

		var deferred = $.Deferred();

		tma.apiStartTap(
				'/mnt/hd/ProgramFiles/' + CLSfilename,
				self.getClsTapParams( filename, doImport, format, silent )
		).done( function( result ) {

			var tap_params = self.decodeClsTapParams( result.tap_params );
			if ( cls_tsr_mode || (tap_params.result == CLS_SUCCESS) ) {

				// wait, until ChannelListSaver is finished
				self.clsFinished( 10 ).done( function() {

					deferred.resolve(result);

				}).fail(function() {

					// CLS timed out
					deferred.reject( CLS_TIMEOUT );

				});

			} else {
				console.log( tap_params );
				// CLS responded with 'error'
				deferred.reject( CLS_ERROR );

			}

		}).fail(function() {

			// starting CLS failed
			deferred.reject( CLS_FAILED );

		});

		return deferred.promise();

	};

	/**
	 * Display the channels obtained from the PVR in the browser window
	 */
	this.show = function() {

		updateStatus.extend( defaultUpdateStatus );

		this.init().done(function() {

			$('#timestamp').hide();

			self.showSticky();

			// check CLS Version
			self.checkClsVersion().done(function() {

				self.showContent();

				self.checkActiveRecordings();

				self.checkSmartEpgVersion();

				// add shortcuts
				self.addShortcuts();

			});

		});

	};

	// help text cache
	var helpText = '';

	/**
	 * reads the module help text
	 */
	this.getModuleHelpText = function() {

		var deferred = $.Deferred();

		if (!helpText) {

			// i18.module.helptext is a filename
			var filename = this.path() + 'lng/' + this.i18.module.info;

			// load help text
			$.get(filename).done(function(data) {

				helpText = data;
				deferred.resolve(helpText);

			});

		} else {

			deferred.resolve(helpText);

		}

		return deferred.promise();

	};

	/**
	 * displays the module help text
	 */
	this.showModuleHelpText = function() {

		if ( this.i18.module.info ) {

			this.getModuleHelpText().done(function(info) {

				var infoWindow = new jkqWindow();
				infoWindow.width = 580;

				infoWindow.showModal();
				infoWindow.setTitle( self.i18.menu_caption );
				infoWindow.closeOnOutsideClick();

				infoWindow.container.find('DIV.jkqWindowContent')
				.append(
						'<div class="cls-help">' +
						info +
						'</div>'
				);

				infoWindow.adjustPos();

			});

		}

	};

	this.showSticky = function() {

		// render template
		$stickyTmpl = JK.template(
				this.templates.head,
				{
					i18:		this.i18,
					channels:	this.channels,
				}
		).appendTo( '.sticky-content' )

		.on('click', '.cls-but-save', function() {
			self.checkClipboard().done(function() {
				var text = self.createClsChannelsFile();
				self.updateDialogue( text );
			});
		})

		.on('click', '.cls-but-export', function() {
			self.checkClipboard().done(function() {
				self.exportChannels();
			});
		})

		.on('click', '.cls-but-help', function() {
			self.showModuleHelpText();
		})

		// import channel file
		.on('change', '#cls-fileinput', function(e) {
			var $form = $(this).parents('FORM');
			if ( e.originalEvent.target.files.length == 1 ) {
				self.importChannels( e.originalEvent.target.files[0] ).done( function( text ) {
					// reset fileinput
					$form.trigger('reset');
					self.updateDialogue( text );
				});
			}
		})
		.on('click', '.cls-but-import', function() {
			$('#cls-fileinput').click();
		})

		;

		$('BODY').click(function() {
			$('.cls-paste-popup').hide();
		});

	};

	/**
	 * Display the channels obtained from the PVR in the browser window
	 */
	this.showContent = function() {

		this.resetChannels();

		//var filename = 'Channels.webc-before.txt';
		var filename = 'channels_' + tma.date('Y-m-d_H-i-s') + '.snap.txt';

//		// 1) remove existing Channels.webc-before.txt
//		self.removeCLSfiles( ['Channels.webc-before.txt', 'Channels.webc-after.txt'] ).done(function() {

			// 2) start ChannelListSaver.tap to create an actual Channels.webc-before.txt
			self.startCLS( filename, CLS_EXPORT, CLS_FMT_TXT, CLS_SILENT ).done(function() {

				// load CLS channels
				self.loadChannels( filename ).done( function() {

					$contentTmpl = JK.template(
							self.templates.content,
							{
								list:		self.list,
								gPrefs:		tma.prefs,
								channels:	self.channels,
								clipboard:	self.clipboard,
								undo:		undoMgr,
								i18:		self.i18,
							}
					).appendTo( 'DIV#content' )

					// Head: delete selected channels from channel list
					.on('click', '.cls-but-delete-list', function() {

						var list = self.channels;
						if ( list.selected ) {
							var channels = list.getSelected();
							self.deleteChannels( channels, list );
						}

					})

					// Head: delete selected channels from clipboard
					.on('click', '.cls-but-delete-clip', function() {

						var list = self.clipboard;
						if ( list.selected ) {
							var channels = list.getSelected();
							self.deleteChannels( channels, list );
						}

					})

					// Head: cut selected channels
					.on('click', '.cls-but-cut', function() {

						var channels = self.channels.getSelected();
						if ( channels.length ) {
							self.cutChannels( channels );
						}

					})

					// Head: insert channels opens options 'insert before/after'
					.on('click', '.cls-but-paste', function(e) {
						e.stopPropagation();

						if ( self.channels.selected ) {
							$('.cls-paste-popup').toggle();
						} else {
							self.insertClipboard( null, false );
						}
					})

					// Head: insert channels before selected channel
					.on('click', '.cls-but-paste-before', function(e) {

						if ( self.channels.selected ) {
							var channel = self.channels.getSelected()[0];
							self.insertClipboard( channel, false );
						}

					})

					// Head: insert channels after selected channel
					.on('click', '.cls-but-paste-after', function(e) {

						if ( self.channels.selected ) {
							var channel = self.channels.getSelected()[0];
							self.insertClipboard( channel, true );
						}

					})

					// Head: text filter
					.on('input', '.cls-text-filter', function() {
						var list = $(this).parents('.cls-channels-frame').is('.cls-channels-frame-clip') ? self.clipboard : self.channels;
						list.filter( $(this).val() );
					})
					.on('click', '.cls-text-filter-reset', function() {
						var list = $(this).parents('.cls-channels-frame').is('.cls-channels-frame-clip') ? self.clipboard : self.channels;
						$(this).siblings('.cls-text-filter').val( '' );
						list.filter( '' );
					})

					// Head: CAS filter
					.on('click', '.cls-cas-filter-fta', function() {
						var list = $(this).parents('.cls-channels-frame').is('.cls-channels-frame-clip') ? self.clipboard : self.channels;
						// toggle bit 0
						var val = list.casFilter ^ 1;
						if ( val === 0) {
							// null is not allowed
							val = 2;
						}
						list.filter( undefined, val );
					})
					.on('click', '.cls-cas-filter-cas', function() {
						var list = $(this).parents('.cls-channels-frame').is('.cls-channels-frame-clip') ? self.clipboard : self.channels;
						// toggle bit 1
						var val = list.casFilter ^ 2;
						if ( val === 0) {
							// null is not allowed
							val = 1;
						}
						list.filter( undefined, val );
					})

					// Head: sort clipboard channels
					.on('click', '.cls-sort-item', function(e) {
						var sortKey = $(this).attr('sort-key');
						var sortDir = $(this).attr('sort-dir');
						self.clipboard.channels.sortBy(sortKey, sortDir);
						$(this).parents( '.cls-sort-box' ).removeClass('cls-sort-box-open');
					})
					.on('touchstart', '.cls-sort-item', function(e) {
						// don't propagate a tap on an item up to box-opener
						e.stopPropagation();
					})

					// open sort options
					.on('touchstart', '.cls-sort', function(e) {
						$('.cls-sort-box').toggleClass('cls-sort-box-open');
					})
					// hover effect with mouse
					.on('mouseenter', '.cls-sort', function(e) {
						$('.cls-sort-box').addClass('cls-sort-box-open');
					})
					.on('mouseleave', '.cls-sort', function(e) {
						$('.cls-sort-box').removeClass('cls-sort-box-open');
					})

					// Head: undo / redo
					.on('click', '.cls-but-undo', function() {
						self.undo();
					})
					.on('click', '.cls-but-redo', function() {
						self.redo();
					})


					// drag&drop
					.on('mousedown', '.cls-channel-ctx', function(e) {

						var $source = $(this);
						if ( $(this).is('[selected="true"]') ) {
							$source = $(this).parents('.cls-channels').find('.cls-channel-ctx[selected="true"]');
						}

						self.dragndrop.start(e, $source, 500);

					})

					// multiselect in channel list
					.on('click', '.cls-channel-ctx', function(e) {

						var list = $(this).parents('.cls-channels-frame').is('.cls-channels-frame-clip') ? self.clipboard : self.channels;
						var clsChannel = list.channel( $(this).attr('channelid') );

						self.setSelection( e, list, clsChannel );

					})

					// delete channel via trash-icon
					.on('click', '.cls-channel-del', function() {

						var list = $(this).parents('.cls-channels-frame').is('.cls-channels-frame-clip') ? self.clipboard : self.channels;

						var clsChannel = list.channel( $(this).parents('.cls-channel-ctx').attr('channelid') );
						if ( clsChannel ) {
							self.deleteChannel(clsChannel, list);
						}

					})

					;

					$('.cls-but-save').prop('disabled', false);
					$('.cls-but-export').prop('disabled', false);
					$('.cls-but-import').prop('disabled', false);

					// adjust vertical size (vScroll in EPG-frame)
					$(window).on('resize.channelsort', function() {

						$contentTmpl.find('.cls-channels').css('height', function() {
							var offs = $(this).offset();
							return $(window).height() - offs.top - 15;
						});

					}).resize();

				});

			}).fail(function( reason ) {

				switch ( reason ) {
					case CLS_FAILED:

						// starting ChannelListSaver.tap failed
						new jkqAlert( {
							title: tma.i18.error.dialog_title,
							text: self.i18.error.cls_start
						}).confirm();
						break;

					case CLS_ERROR:

						// ChannelListSaver.tap did not return with result=success
						new jkqAlert( {
							title: tma.i18.error.dialog_title,
							text: self.i18.error.cls_failed
						}).confirm();
						break;

					case CLS_TIMEOUT:

						// finishing ChannelListSaver.tap failed
						new jkqAlert( {
							title: tma.i18.error.dialog_title,
							text: self.i18.error.cls_running
						}).confirm();
						break;

				}

			});

//		});

	};

	this.setSelection = function( e, list, clsChannel ) {

		if (e.ctrlKey) {
			list.toggleSelected( clsChannel );
		} else if (e.shiftKey && (list.recentSelected)) {

			if ( clsChannel !== list.recentSelected ) {
				var inRange = false;
				for ( var i = 0, len = list.channels.length; i < len; i++ ) {
					var ch = list.channels[i];
					if ( ch.channel.chtype === list.chtype ) {
						// channel is visible

						if ( ch === list.recentSelected ) {
							inRange = !inRange;
						} else if ( ch === clsChannel ) {
							list.toggleSelected( ch );
							inRange = !inRange;
						} else if (inRange) {
							list.toggleSelected( ch );
						}
					}
				}
			}

		} else if ( clsChannel ){
			var wasSelected = clsChannel.selected && (list.selected == 1);
			list.deselectAll();
			if ( !wasSelected ) {
				// select channel
				list.toggleSelected( clsChannel );
			}
		}

		// store recent selection
		if ( clsChannel && clsChannel.selected ) {
			list.recentSelected = clsChannel;
		}

		if ( list.selected === 0 ) {
			// if no channel is delected, remove recentSelected
			list.recentSelected = null;
		}

	};

	/**
	 * delete a channel from list (channel list or clipboard)
	 *
	 * @param {ClsChannel}	clsChannel
	 * @param {ClsChannels} list		channel list or clipboard
	 */
	this.deleteChannel = function( clsChannel, list ) {

		new jkqDialog({
			title:			this.i18.dialog.del.title,
			text:			this.i18.dialog.del.text.format({channel: clsChannel.channel.servicename}),
			autoSkipText:	tma.i18.respond_auto_skip,
			autoSkipButton:	'clsDeleteChannel',
			buttons:		{
				jkqDialogCancel:	tma.i18.respond_cancel,
				clsDeleteChannel:	this.i18.button.del.text
			}
		}).confirm(function(button) {

			if (button == 'clsDeleteChannel') {

				var undo = list.deleteChannel( clsChannel );
				if (list === self.channels) {
					list.renumber();
				}

				// add to undo stack
				var undoGroup = [ undo ];
				self.undoAdd( undoGroup );

			}

		});

	};

	/**
	 * delete a set of channels from list (channel list or clipboard)
	 *
	 * @param {Array}		channels	Array of ClsChannels
	 * @param {ClsChannels}	list		channel list or clipboard
	 */
	this.deleteChannels = function( channels, list ) {

		new jkqDialog({
			title:			this.i18.dialog.delsel.title,
			text:			this.i18.dialog.delsel.text,
			autoSkipText:	tma.i18.respond_auto_skip,
			autoSkipButton:	'clsDeleteChannels',
			buttons:		{
				jkqDialogCancel:	tma.i18.respond_cancel,
				clsDeleteChannels:	this.i18.button.del.text
			}
		}).confirm(function(button) {
			var i = 0;
			var len = 0;

			if (button == 'clsDeleteChannels') {
				var undoGroup = [];
				for ( i = 0, len = channels.length; i < len; i++ ) {
					var undo = list.deleteChannel( channels[i] );
					undoGroup.push( undo );
				}
				if (list === self.channels) {
					list.renumber();
				}

				// add to undo stack
				self.undoAdd( undoGroup );
			}

		});

	};

	/**
	 * remove a set of channels from channel list and add them to clipboard
	 *
	 * @param {Array}	channels	Array of ClsChannels
	 */
	this.cutChannels = function( channels ) {

		var undoGroup = [];
		for ( var i = 0, len = channels.length; i < len; i++ ) {
			// remove channel from list
			undoGroup.push( this.channels.deleteChannel( channels[i] ) );
			// add to clipboard
			undoGroup.push( this.clipboard.insert( channels[i] ) );
		}

		// re-number channels
		this.channels.renumber();

		// add to undo stack
		this.undoAdd( undoGroup );

	};

	/**
	 * insert channels from clipboard before/after channel
	 *
	 * @param {ClsChannel}	channel
	 * @param {Boolean}		insertAfter		insert after channel?
	 */
	this.insertClipboard = function( channel, insertAfter ) {
		var MODE_APPEND = 999999;

		var i = 0;
		var len = 0;
		var targetIndex = MODE_APPEND;

		insertAfter = (insertAfter) ? 1 : 0;

		if ( channel instanceof ClsChannel ) {
			targetIndex = self.channels.channels.findIndex( function( ch ) {
				return channel === ch;
			});
		}

		if ( targetIndex > -1) {

			if ( !this.clipboard.selected ) {
				// select channels to insert
				this.clipboard.selectAll();
			}

			var channels = this.clipboard.getSelected();
			if ( channels.length ) {

				var undoGroup = [];

				this.channels.deselectAll();

				for( i = 0, len = channels.length; i < len; i++ ) {
					// remove channel from clipboard
					undoGroup.push( this.clipboard.deleteChannel( channels[i] ) );
					// add to list
					if ( targetIndex == MODE_APPEND ) {
						undoGroup.push( this.channels.insert( channels[i] ) );
					} else {
						undoGroup.push( this.channels.insert( channels[i], targetIndex + insertAfter + i ) );
					}
				}

				// re-number channels
				this.channels.renumber();

				// add to undo stack
				this.undoAdd( undoGroup );

			}
		}

	};

	/**
	 * Drag&Drop Objekt
	 */
	this.dragndrop = new jkqDragDrop(
			// Drop-Targets
			'.cls-channel-ctx:visible,.cls-clipboard',
			//'.cls-channel-ctx:visible',
			// onDragDrop
			function(target, source) {
				var i = 0;
				var len = 0;
				var insertBefore = 0;
				var insertAfter  = 0;
				var targetList;
				var targetChannel;
				var targetIndex;
				var sourceList = $(source).first().parents('.cls-channels-frame').is('.cls-channels-frame-clip') ? self.clipboard : self.channels;

				if ( $(target).is('.cls-clipboard') ) {
					// special case: drag over clipboard
					if ( sourceList === self.channels ) {
						// allowed, if drag from list
						insertAfter  = 1;
						targetList = self.clipboard;
					} else {
						// not allowed, if drag from clipboard
						return;
					}
				} else {

					// test insert permission and evaluate insertion before/after
					var sourceFirstChannel = sourceList.channel( $(source).first().attr('channelid') );
					var sourceIndexFirst = sourceList.channels.indexOf( sourceFirstChannel );
					var sourceLastChannel = sourceList.channel( $(source).last().attr('channelid') );
					var sourceIndexLast = sourceList.channels.indexOf( sourceLastChannel );

					targetList = $(target).parents('.cls-channels-frame').is('.cls-channels-frame-clip') ? self.clipboard : self.channels;
					targetChannel = targetList.channel( $(target).attr('channelid') );
					targetIndex = targetList.channels.indexOf( targetChannel );

					if (
							(sourceList !== targetList) ||
							(sourceIndexFirst > targetIndex)
					) {
						insertBefore = 1;
					} else if (sourceIndexLast < targetIndex) {
						insertAfter  = 1;
					}

				}

				if ( insertBefore || insertAfter ) {

					var channels = [];

					// collect dragged channels
					$(source).each(function(i, srcChannelCtx) {
						var clsChannel = sourceList.channel( $(srcChannelCtx).attr('channelid') );
						if ( clsChannel ) {
							channels.push( clsChannel );
						}
					});

					var undoGroup = [];

					// remove channels from source list
					for( i = 0, len = channels.length; i < len; i++ ) {
						undoGroup.push( sourceList.deleteChannel( channels[i] ) );
					}

					if ( targetChannel ) {
						// update index of target channel
						targetIndex = targetList.channels.indexOf( targetChannel );

						// insert channels into target list
						for( i = 0, len = channels.length; i < len; i++ ) {
							undoGroup.push( targetList.insert( channels[i], targetIndex + insertAfter + i ) );
						}
					} else {

						// add channels to target list (clipboard)
						for( i = 0, len = channels.length; i < len; i++ ) {
							undoGroup.push( targetList.insert( channels[i] ) );
						}

					}

					// re-number channels
					self.channels.renumber();

					// add to undo stack
					self.undoAdd( undoGroup );

				}

			},
			// onDragOver
			function(target, source, xOffs, yOffs) {

				var sourceList = $(source).first().parents('.cls-channels-frame').is('.cls-channels-frame-clip') ? self.clipboard : self.channels;
				if ( $(target).hasClass('cls-clipboard') ) {
					// special case: drag over clipboard
					if ( sourceList === self.channels ) {
						// allowed, if drag from list
						$(target).addClass('cls-insertafter');
						this.$container.addClass('jkDropable');
						return true;
					} else {
						// not allowed, if drag from clipboard
						return false;
					}
				}
				var sourceFirstChannel = sourceList.channel( $(source).first().attr('channelid') );
				var sourceIndexFirst = sourceList.channels.indexOf( sourceFirstChannel );
				var sourceLastChannel = sourceList.channel( $(source).last().attr('channelid') );
				var sourceIndexLast = sourceList.channels.indexOf( sourceLastChannel );

				var targetList = $(target).parents('.cls-channels-frame').is('.cls-channels-frame-clip') ? self.clipboard : self.channels;
				var targetChannel = targetList.channel( $(target).attr('channelid') );
				var targetIndex = targetList.channels.indexOf( targetChannel );

				if (
						(sourceList !== targetList) ||
						(sourceIndexFirst > targetIndex)
				) {
					$(target).addClass('cls-insertbefore');
					this.$container.addClass('jkDropable');
					return true;
				} else if (sourceIndexLast < targetIndex) {
					$(target).addClass('cls-insertafter');
					this.$container.addClass('jkDropable');
					return true;
				}
			},
			// onDragOut
			function(target) {
				this.$container.removeClass('jkDropable');
				$(target).removeClass('cls-insertbefore cls-insertafter');
			}

	);
	// override drag start
	this.dragndrop.onDragStart = function(e, drag_source) {
		this.defaultDragStart(e, drag_source);

		// hack: set clipboard as very last drop target
		this.drop_targets.sort( function( a, b ) {
			return ( $(a).hasClass('cls-clipboard') ? 1 : -1);
		});
	};

	this.undoAdd = function( undoGroup ) {

		if ( undoGroup.length ) {

			// add to undo stack
			if ( undoMgr.undoPtr < undoMgr.stack.length ) {
				undoMgr.stack.length = undoMgr.undoPtr;
			}
			undoMgr.stack.push( undoGroup );
			undoMgr.undoPtr = undoMgr.stack.length;
			undoMgr.redoPtr = 0;

		}

	};

	this.undo = function() {
		if ( undoMgr.undoPtr > 0 ) {
			var undoGroup = undoMgr.stack[ --undoMgr.undoPtr ];
			undoMgr.redoPtr++;

			var renumber = false;
			for ( var i = 0, len = undoGroup.length; i < len; i++ ) {
				var undo = undoGroup[i];
				switch (undo.action) {

				case Undo.DELETE:
					undo.list.insert( undo.channel, undo.position );
					if ( undo.list === this.channels ) {
						renumber = true;
					}
					break;

				case Undo.INSERT:
					undo.list.deleteChannel( undo.channel );
					if ( undo.list === this.channels ) {
						renumber = true;
					}
					break;

				}
			}
			if ( renumber ) {
				this.channels.renumber();
			}
		}
	};

	this.redo = function() {
		if ( undoMgr.redoPtr > 0 ) {
			var redoGroup = undoMgr.stack[ undoMgr.undoPtr++ ];
			undoMgr.redoPtr--;

			var renumber = false;
			for ( var i = 0, len = redoGroup.length; i < len; i++ ) {
				var redo = redoGroup[i];
				switch (redo.action) {

				case Undo.INSERT:
					redo.list.insert( redo.channel, redo.position );
					if ( redo.list === this.channels ) {
						renumber = true;
					}
					break;

				case Undo.DELETE:
					redo.list.deleteChannel( redo.channel );
					if ( redo.list === this.channels ) {
						renumber = true;
					}
					break;

				}
			}
			if ( renumber ) {
				this.channels.renumber();
			}
		}
	};

	this.checkActiveRecordings = function() {

		var activeRecs = tma.status.activeRecordingCount();
		if ( activeRecs > 0 ) {

			// warning of active recordings
			new jkqAlert( {
				title: this.i18.warning.dialog_title,
				text: this.i18.warning.active_recs.format( activeRecs )
			}).confirm();

		}

	};

	/**
	 * returns a float representation of a TAP version string
	 */
	function parseVersion( verStr ) {

		verStr = verStr.replace(/^[a-z ]+/i, '');
		var version = parseFloat( verStr );
		var vMain = version;
		if ( verStr.match(/RC/) ) {
			var vRC = 1 + parseInt( verStr.replace(/.*RC[ ]*/, '') );
			version = vMain - 0.01 + vRC / 10000;
			//console.log( 'Version', vMain, 'RC', vRC, '=>', version );
		}
		if ( verStr.match( /[0-9][ ]*[a-z]$/) ) {
			var vSub = verStr.charCodeAt(verStr.length - 1) - 'a'.charCodeAt(0) + 1;
			version = vMain + vSub / 10000;
			console.log( 'Version', vMain, 'Sub', vSub, '=>', version );
		}

		return version;

	}

	/**
	 * checks, if ChannelListSaver is available and its version is recent enough
	 *
	 * @return {jQuery.Promise}
	 */
	this.checkClsVersion = function() {

		var deferred = $.Deferred();

		var files = new TmaFiles();
		tma.apiGetFiles( '/mnt/hd/ProgramFiles', files ).done(function() {

			//var clsTap = files.find( 'filename', 'ChannelListSaver.tap' );
			//var clsTap = files.find( 'tap.tapid_hex', '2a0a0002' );
			var clsTap = files.files.find( function( file) {
					return ( file.tap && file.tap.tapid == CLStapid );
			});
			if (clsTap) {

				CLSfilename = clsTap.name;
				var version = parseVersion( clsTap.tap.version );
				var minVersion = parseVersion( Min_CLS_Version );
				if (version >= minVersion) {

					// CLS is valid
					deferred.resolve();

				} else {

					// ChannelListSaver.tap version is to low
					$('.cls-but-save').prop('disabled', true);
					new jkqAlert( {
						title: tma.i18.error.dialog_title,
						text: self.i18.error.cls_version.format(clsTap.tap.version, Min_CLS_Version)
					}).confirm( function() {
						deferred.reject();
					});

				}

			} else {

				// searching ChannelListSaver.tap failed
				$('.cls-but-save').prop('disabled', true);
				new jkqAlert( {
					title: tma.i18.error.dialog_title,
					text: self.i18.error.cls_not_found
				}).confirm( function() {
					deferred.reject();
				});

			}

		}).fail( function() {
			deferred.reject();
		});

		return deferred.promise();

	};

	/**
	 * checks, if SmartEPG version is recent enough
	 *
	 * @return {jQuery.Promise}
	 */
	this.checkSmartEpgVersion = function() {

		var deferred = $.Deferred();

		// test, if SmartEPG if running
		var se = tma.taps.find( 0x8E0A424E );

		if ( !se ) {

			// SmartEPG isn't active, so do not care about
			deferred.resolve();

		} else {

			var version = parseVersion( se.version );
			var minVersion = parseVersion( Min_SE_Version );
			if (version >= minVersion) {

				// SmartEPG version is valid
				deferred.resolve();

			} else {

				// SmartEPG.tap version is to low
				//$('.cls-but-save').prop('disabled', true);
				new jkqAlert( {
					title: tma.i18.error.dialog_title,
					text: self.i18.error.se_version.format(se.version, Min_SE_Version)
				}).confirm( function() {
					deferred.reject();
				});

			}

		}

		return deferred.promise();

	};

	/**
	 * remove files from Settings/ChannelListSaver/
	 *
	 * @param {Array}	files
	 * @returns {jQueryDeferred}
	 */
	this.removeCLSfiles = function( files ) {

		var deferred = $.Deferred();
		var pipe = deferred;

		$.each(files, function(i, filename) {

			pipe = pipe.then(function() {

				var pipeDeferred = $.Deferred();

				tma.apiDeleteFile( CLSdirectory, filename).done(function( result ) {
					if (result.mpg_status === 0) {
						console.log('"' + filename + '" removed');
					} else {
						console.log('"' + filename + '" not found');
					}
				}).always(function() {
					pipeDeferred.resolve();
				});

				return pipeDeferred.promise();

			});

		});

		deferred.resolve();

		return pipe.promise().done(function() {

		});

	};

	/**
	 * wait until ChannelListSaver has stopped
	 *
	 * @param {Number}	timeout		in seconds
	 * @returns jQueryDeferred
	 */
	this.clsFinished = function( timeout ) {

		var deferred = $.Deferred();
		var loop = timeout;

		if (tma.config.mock ) {
			deferred.resolve();
		} else {
			var timer = window.setInterval(
					function() {

						tma.apiGetTaps().done( function() {

							if (loop > 0) {
								loop--;
								if ( tma.taps.find( CLStapid ) ) {
									if (loop === 0) {
										// timeout error: CLS is still running
										window.clearInterval( timer );
										deferred.reject();
									}
								} else {
									// CLS has stopped
									window.clearInterval( timer );
									deferred.resolve();
									loop = 0;
								}
							}
						});

					},
					2000
			);
		}

		return deferred.promise();

	};

	/**
	 * helper for CLS channels.txt scan
	 * searches the last semicolon in line
	 *
	 * @param {Uint8Array}	line
	 * @returns {Number}
	 */
	function ui8FindLastSemicolon( line ) {

		var ui8Semicolon = ';'.charCodeAt(0);

		var result = -1;
		for ( var i = line.length - 1; i >= 0 ; i--) {
			if (line[i] == ui8Semicolon) {
				result = i;
				break;
			}
		}

		return result;

	}

	/**
	 * helper for CLS channels.txt scan
	 * searches the nth semicolon in line
	 *
	 * @param {Uint8Array}	line
	 * @param {Number}		n
	 * @returns {Number}
	 */
	function ui8FindNthSemicolon( line, n ) {

		var ui8Semicolon = ';'.charCodeAt(0);

		var result = -1;
		for ( var i = 0, cnt = 0, len = line.length; i < len ; i++) {
			if (line[i] == ui8Semicolon) {
				cnt++;
				if ( cnt === n ) {
					result = i;
					break;
				}
			}
		}

		return result;

	}

	/**
	 * helper for CLS channels.txt scan
	 * parses a number from a Uint8Array starting from position
	 *
	 * @param {Uint8Array}	ui8
	 * @param {Number}		position
	 * @returns {Number}
	 */
	function ui8ReadInt ( ui8, position ) {

		var i = position;
		var len = ui8.length;

		// read number
		var s = '';
		// trim spaces at start
		while ( (i < len) && (ui8[i] == 32) ) {
			i++;
		}

		while ( (i < len) && (ui8[i] >= 48) && (ui8[i] <= 57) ) {
			s += (ui8[i] - 48).toString();
			i++;
		}
		return parseInt( s );

	}

	/**
	 * helper for CLS channels.txt scan
	 * searches the complete text for '/caption/=/number/ and returns the number
	 *
	 * @param {String}	caption
	 * @returns {Number}
	 */
	function ui8ReadIntVal ( caption ) {

		var ui8Label	= Uint8Array.createFromString( caption + '=' );

		var i = cls.text.uInt8IndexOf( ui8Label );
		if ( i < 0 ) return -1;
		// read number
		i += ui8Label.length;
		return ui8ReadInt(cls.text, i);

	}

	this.resetChannels = function() {

		// reset
		updateStatus.extend( defaultUpdateStatus );
		this.channels.reset();
		this.clipboard.reset();
		// init channel list
		this.channels.favGroup = '__tv';

	};

	/**
	 * parse CLS text file from Uint8Array in cls.text
	 */
	this.parseChannels = function() {

		// some useful constants
		var ui8CRLF			= Uint8Array.createFromString( CRLF );
		var ui8Satellites	= Uint8Array.createFromString( '[Satellites]' + CRLF );
		var ui8TVServices	= Uint8Array.createFromString( '[TVServices]' + CRLF );
		var ui8RadioServices= Uint8Array.createFromString( '[RadioServices]' + CRLF );
		var ui8Favourites	= Uint8Array.createFromString( '[Favorites]' + CRLF );

		var clsChannel;
		var linestart = 0;
		var i, p, groupNr, j, len, svcId;

		// read content parameters
		cls.FileVersion		= ui8ReadIntVal( 'FileVersion' );
		cls.SystemType		= ui8ReadIntVal( 'SystemType' );
		cls.UTF8System		= ui8ReadIntVal( 'UTF8System' );
		cls.NrSatellites	= ui8ReadIntVal( 'NrSatellites' );
		cls.NrTransponders	= ui8ReadIntVal( 'NrTransponders' );
		cls.NrFavsPerGroup	= ui8ReadIntVal( 'NrFavsPerGroup' );


		// find satellites start
		i = cls.text.uInt8IndexOf( ui8Satellites );
		if ( i < 0 ) return;
		cls.slices.beforeTV.start = i;

		// find TV channels start
		i = cls.text.uInt8IndexOf( ui8TVServices );
		if ( i < 0 ) return;

		// find next line end (skip column header)
		i = cls.text.uInt8IndexOf( ui8CRLF, i + ui8TVServices.length );
		if ( i < 0 ) return;
		// remember slice before first channel line
		cls.slices.beforeTV.end   = i + 2;

		// copy channel lines
		linestart = i + 2;
		i = cls.text.uInt8IndexOf( ui8CRLF, linestart );
		var chnum = 0;
		var line = '';
		var channel;
		while ( (i >= 0) && (i > linestart) ) {
			// text line including CRLF
			line = cls.text.slice(linestart, i + 2);
			// find serviceID in line
			//p = ui8FindNthSemicolon( line, 7 );
			//svcId = ui8ReadInt( line, p + 1 );
			// find channel by service id
			//channel = findCHannelByServiceId( svcId );
			channel = tma.channels.channelByTypeNum(TmaChannel.TYPE_TV, chnum);
			if (channel) {
				// workaround for CLS 0.8 bug with empty channel names
				if ( line[6] == 32 ) {
					line[6] = '-'.charCodeAt(0);
				}
				// create sortable channel object
				clsChannel = new ClsChannel({
					clsLine		: line,			// bind channel line to channel object
					number		: chnum + 1,
					originalNr	: channel.chnum + 1,
					selected	: false,
					channel		: channel,
				});
				self.channels.channels.push( clsChannel );
			}
			linestart = i + 2;
			i = cls.text.uInt8IndexOf( ui8CRLF, linestart );
			chnum++;
		}

		// find radio channels start
		i = cls.text.uInt8IndexOf( ui8RadioServices, linestart );
		if ( i < 0 ) return;
		// find next line end (skip column header)
		i = cls.text.uInt8IndexOf( ui8CRLF, i + ui8RadioServices.length );
		if ( i < 0 ) return;
		// remember slice before first channel line
		cls.slices.beforeRadio.start = linestart;
		cls.slices.beforeRadio.end   = i + 2;

		// copy channel lines
		linestart = i + 2;
		i = cls.text.uInt8IndexOf( ui8CRLF, linestart );
		chnum = 0;
		while ( (i >= 0) && (i > linestart) ) {
			// text line including CRLF
			line = cls.text.slice(linestart, i + 2);
			// find serviceID in line
			//p = ui8FindNthSemicolon( line, 7 );
			//svcId = ui8ReadInt( line, p + 1 );
			// find channel by service id
			//channel = findCHannelByServiceId( svcId );
			channel = tma.channels.channelByTypeNum(TmaChannel.TYPE_RADIO, chnum);
			if (channel) {
				// workaround for CLS 0.8 bug with empty channel names
				if ( line[6] == 32 ) {
					line[6] = '-'.charCodeAt(0);
				}
				// create sortable channel object
				clsChannel = new ClsChannel({
					clsLine		: line,			// bind channel line to channel object
					number		: chnum + 1,
					originalNr	: channel.chnum + 1,
					selected	: false,
					channel		: channel,
				});
				self.channels.channels.push( clsChannel );
			}
			linestart = i + 2;
			i = cls.text.uInt8IndexOf( ui8CRLF, linestart );
			chnum++;
		}

		// find favourites start
		i = cls.text.uInt8IndexOf( ui8Favourites, linestart );
		if ( i < 0 ) return;
		// find next line end (skip column header)
		i = cls.text.uInt8IndexOf( ui8CRLF, i + ui8Favourites.length );
		if ( i < 0 ) return;
		// remember slice before first favourite line
		cls.slices.beforeFav.start = linestart;
		cls.slices.beforeFav.end   = i + 2;

		// copy favourites lines
		linestart = i + 2;
		i = cls.text.uInt8IndexOf( ui8CRLF, linestart );
		p = 0;
		groupNr = 0;
		j = 0;
		len = 0;
		while ( (i >= 0) && (i > linestart) ) {
			line = cls.text.slice(linestart, i + 2);
			p = ui8FindLastSemicolon( line );
			if (p >= 0) {
				tma.favourites.groups[groupNr].clsLineStart = line.slice(0, p + 1);
				tma.favourites.groups[groupNr].clsIsTV = ( line[p - 1] == 'T'.charCodeAt(0) );
			}
			linestart = i + 2;
			i = cls.text.uInt8IndexOf( ui8CRLF, linestart );
			groupNr++;
		}

	};

	/**
	 * reads a CLS channels text file into a Uint8Array and parses the content
	 */
	this.loadChannels = function( filename ) {

		/**
		 * this construct allows us to switch to fail(),
		 * if the api returns an error_code <br>
		 * see: return deferred.promise().fail()
		 * http://stackoverflow.com/a/5112734
		 */
		var deferred = $.Deferred();

		try {

			var xhr = new XMLHttpRequest();
			xhr.open('GET', CLSdirectory + '/' + filename, true);
			xhr.responseType = 'arraybuffer';

			xhr.onload = function(e) {

				if ( this.status == 200 ) {

					cls.text = new Uint8Array(this.response);

					self.parseChannels();

					deferred.resolve(this);

				} else {
					deferred.reject( this, this.status, this.statusText );
				}

			};
			xhr.onerror = function(e) {
				deferred.reject(this, 'xhr-error', 'xhr-error');
			};

			xhr.send();

		} catch (e) {
			deferred.reject(null, e.name, e.message);
		}

		return deferred.promise().fail(function(xhr, errorType, error) {

			tma.log({
				title: errorType + " in channelsort.loadChannels()",
				details: {
					error: error,
				}
			});

		});

	};

	/**
	 * creates the text file for ChannelListSaver
	 *
	 * @returns {Uint8Array}
	 */
	this.createClsChannelsFile = function() {
		var i = 0,
			j = 0,
			len = 0,
			pos = 0,
			channel,
			clsChannel,
			chnum = 0,
			chnumStr = '',
			groupNr = 0,
			favBytes = 0,
			group;

		// prepare favourites
		for ( groupNr = 0; groupNr < tma.favourites.groups.length; groupNr++) {

			chnumStr = '';
			var chnums = [];
			group = tma.favourites.groups[groupNr];
			if (group.channels.length > 0) {
				for( j = 0; j < group.channels.length; j++) {
					channel = group.channels[j];
					clsChannel = this.channels.channel( channel.logoid );
					if ( clsChannel ) {
						chnums.push( clsChannel.number - 1 );
					}
				}
				chnumStr = '  ' + chnums.join(', ');
			}
			chnumStr += CRLF;

			// create new byte array for fav-line
			var line = new Uint8Array(group.clsLineStart.length + chnumStr.length);
			// copy group name and T/R type
			pos = 0;
			for( i = 0, len = group.clsLineStart.length; i < len; i++, pos++ ) {
				line[pos] = group.clsLineStart[i];
			}
			// copy channel numbers
			var ui8ChnumStr = Uint8Array.createFromString( chnumStr );
			for( i = 0, len = ui8ChnumStr.length; i < len; i++, pos++ ) {
				line[pos] = ui8ChnumStr[i];
			}

			group.clsLine = line;
			favBytes += line.length;

		}

		// Nr. of bytes used by TV channels
		var tvBytes = 0;

		// get TV channels
		var tvchannels = [];
		for ( i = 0, len = this.channels.channels.length; i < len; i++ ) {
			clsChannel = this.channels.channels[i];
			if ( clsChannel.channel.chtype === TmaChannel.TYPE_TV ) {
				tvBytes += clsChannel.clsLine.length;
				tvchannels.push( clsChannel );
			}
		}

		// Nr. of bytes used by Radio channels
		var radioBytes = 0;

		// get Radio channels
		var radiochannels = [];
		for ( i = 0, len = this.channels.channels.length; i < len; i++ ) {
			clsChannel = this.channels.channels[i];
			if ( clsChannel.channel.chtype === TmaChannel.TYPE_RADIO ) {
				tvBytes += clsChannel.clsLine.length;
				radiochannels.push( clsChannel );
			}
		}

		// prepare content parameters
		var params  = '[ChannelListSaver]' + CRLF;
			params += 'FileVersion='	+ cls.FileVersion + CRLF;
			params += 'FileSize=0000000000' + CRLF;
			params += 'NrOfLines=1111111111' + CRLF;
			params += 'SystemType='		+ cls.SystemType + CRLF;
			params += 'UTF8System='		+ cls.UTF8System + CRLF;
			params += 'NrSatellites='	+ cls.NrSatellites + CRLF;
			params += 'NrTransponders=' + cls.NrTransponders + CRLF;
			params += 'NrTVServices='	+ tvchannels.length + CRLF;
			params += 'NrRadioServices='+ radiochannels.length + CRLF;
			params += 'NrFavGroups='	+ tma.favourites.groups.length + CRLF;
			params += 'NrFavsPerGroup=' + cls.NrFavsPerGroup + CRLF;
			params += CRLF;


		var filesize =	params.length +
						cls.slices.beforeTV.end - cls.slices.beforeTV.start +		// sat, transp., tv-header
						tvBytes +
						cls.slices.beforeRadio.end - cls.slices.beforeRadio.start +	// radio-header
						radioBytes +
						cls.slices.beforeFav.end - cls.slices.beforeFav.start + 	// fav-header
						favBytes;
		var nrOfLines = 26 + cls.NrSatellites * 3 + cls.NrTransponders + tvchannels.length + radiochannels.length + tma.favourites.groups.length;

		// replace filesize
		var filesizeStr = ('0000000000' + filesize).slice(-10);
		params = params.replace(/0000000000/, filesizeStr);

		// replace NrOfLines
		var nrOfLinesStr = ('0000000000' + nrOfLines).slice(-10);
		params = params.replace(/1111111111/, nrOfLinesStr);

		var text = new Uint8Array(filesize);
		pos = 0;

		// copy content params
		for( i = 0, len = params.length; i < len; i++, pos++ ) {
			text[pos] = params.charCodeAt(i);
		}

		// copy satellites, transponders and TV header
		for( i = cls.slices.beforeTV.start, len = cls.slices.beforeTV.end; i < len; i++, pos++ ) {
			text[pos] = cls.text[i];
		}

		// copy TV channels
		for( j = 0; j < tvchannels.length; j++ ) {
			channel = tvchannels[j];
			// replace channel number
			chnumStr = ('    ' + (channel.number - 1)).slice(-4);
			for( i = 0; i < 4; i++ ) {
				channel.clsLine[i] = chnumStr.charCodeAt(i);
			}
			for( i = 0, len = channel.clsLine.length; i < len; i++, pos++ ) {
				text[pos] = channel.clsLine[i];
			}
		}

		// copy Radio header
		for( i = cls.slices.beforeRadio.start, len = cls.slices.beforeRadio.end; i < len; i++, pos++ ) {
			text[pos] = cls.text[i];
		}

		// copy Radio channels
		for( j = 0; j < radiochannels.length; j++ ) {
			/** @returns {ClsChannel} */
			channel = radiochannels[j];
			// replace channel number
			chnumStr = ('    ' + (channel.number - 1)).slice(-4);
			for( i = 0; i < 4; i++ ) {
				channel.clsLine[i] = chnumStr.charCodeAt(i);
			}
			for( i = 0, len = channel.clsLine.length; i < len; i++, pos++ ) {
				text[pos] = channel.clsLine[i];
			}
		}

		// copy Fav header
		for( i = cls.slices.beforeFav.start, len = cls.slices.beforeFav.end; i < len; i++, pos++ ) {
			text[pos] = cls.text[i];
		}

		// copy favourites
		for( groupNr = 0; groupNr < tma.favourites.groups.length; groupNr++ ) {
			group = tma.favourites.groups[groupNr];
			for( i = 0, len = group.clsLine.length; i < len; i++, pos++ ) {
				text[pos] = group.clsLine[i];
			}
		}

		return text;

	};

	/**
	 * creates the text file for ChannelListSaver,
	 * uploads the file and checks the upload
	 *
	 * @param {string}		filename	desired filename without path
	 * @param {Uint8Array}	text
	 * @returns {jQueryDeferred}		if upload succeded, the done-callback parameter
	 * 									contains result object with property filesize
	 */
	this.saveChannels = function( filename, text ) {

		var filesize = text.length;

		/**
		 * this construct allows us to switch to fail(),
		 * if the api returns an error_code <br>
		 * see: return deferred.promise().fail()
		 * http://stackoverflow.com/a/5112734
		 */
		var deferred = $.Deferred();

		//var filename = CLSdirectory + "/Channels.webc-after.txt";
		var url = '/api?function=files&action=store!&filename=' + CLSdirectory + '/' + filename + '&';

		try {

			var xhr = new XMLHttpRequest();
			xhr.responseType = 'json';
			xhr.open('POST', url, true);
			xhr.onload = function(e) {
				if ( this.response ) {
					if (this.response.errno === 0) {
						var response = $.extend({}, this.response, {filesize: filesize});
						deferred.resolve(response);
					} else {
						deferred.reject(this, 'tma-error', 'Upload channels: errno = ' + this.response.errno);
					}
				} else {
					deferred.reject(this, 'tma-error', 'Upload channels: response is undefined');
				}
			};
			xhr.onerror = function(e) {
				deferred.reject(this, 'xhr-error', 'xhr-error');
			};

			xhr.send(text.buffer);

		} catch (e) {
			deferred.reject(null, e.name, e.message);
		}

		return deferred.promise().fail(function(xhr, errorType, error) {

			tma.log({
				title: errorType + " in channelsort.saveChannels()",
				details: {
					error: error,
				}
			});

		});

	};

	/**
	 * stop SmartEPG, if param SE is instance of TmaTap
	 *
	 * @param {TmaTap}	SE
	 * @returns {jQueryDeferred}
	 */
	this.stopSE = function( SE ) {

		if ( SE instanceof TmaTap ) {
			// stop SmartEPG
			return tma.apiStopTap( SE );
		} else {
			// return immediately
			return $.Deferred().resolve().promise();
		}

	};

	/**
	 * start SmartEPG, if SE is instance of TmaTap
	 *
	 * @param {TmaTap}	SE
	 * @returns {jQueryDeferred}
	 */
	this.startSE = function( SE ) {

		var deferred = $.Deferred();

		if ( SE instanceof TmaTap ) {

			// inform SmartEPG to update channel list:
			// save "RescanChannelList"
			var fileRescanChannelList = '/mnt/hd/ProgramFiles/Settings/SmartEPG_TMS/RescanChannelList';
			tma.apiSaveFile( fileRescanChannelList, '' ).done(function() {
				// start SmartEPG again
				tma.apiStartTap( SE.file_name ).done(function() {
					deferred.resolve();
				}).fail(function() {
					// something went wrong
					deferred.reject();
				});
			}).fail(function() {
				// something went wrong
				deferred.reject();
			});

		} else {
			// no SmartEPG: return immediately
			deferred.resolve();
		}

		return deferred.promise();

	};

	// !!! cls now handles timers
//	/**
//	 * if SE is instance of TmaTap, wait until timers are recreated by SE, else recreate timers
//	 *
//	 * @param {TmaTap}	SE
//	 * @param {JKarray}	prevTimers
//	 * @returns {jQueryDeferred}
//	 */
//	this.waitForOrRecreateTimers = function( SE, prevTimers ) {
//
//		var deferred = $.Deferred();
//		var i, len, timer;
//
//		var pipe = deferred;
//
//		function recreateTimer( timer ) {
//
//			timer.index = -1;
//			timer.modReset();
//			pipe = pipe.always(function() {
//				return tma.apiTimerUpdate( timer ).done(function() {
//					updateStatus.postTimersCnt++;
//				});
//			});
//
//		}
//
//		if (prevTimers.length === 0) {
//
//			// nothing to do => return immediately
//			return deferred.resolve().promise();
//
//		} else if ( SE instanceof TmaTap ) {
//
//			// wait until SE has refreshed the timers
//			var loop = 20;
//
//			if (tma.config.mock ) {
//				deferred.resolve();
//			} else {
//				var tim = window.setInterval(
//						function() {
//							if (!loop) return;
//							tma.timers.load().done( function() {
//
//								if (loop > 0) {
//									if (tma.timers.timers.length > 0) {
//										// SE has recreated the timers
//										updateStatus.postTimersCnt = tma.timers.timers.length;
//										window.clearInterval( tim );
//										// only once
//										loop = 0;
//
//										deferred.resolve();
//									} else {
//										loop--;
//										if (loop === 0) {
//											// timeout error: still no timers found
//											window.clearInterval( tim );
//											deferred.reject();
//										}
//									}
//								}
//							});
//
//						},
//						1000
//				);
//			}
//
//			return deferred.promise();
//
//		} else {
//
//			// recreate timer list
//
//			for ( i = 0, len = prevTimers.length; i < len; i++ ) {
//
//				if (prevTimers[i].modChannel() instanceof TmaChannel) {
//					recreateTimer( prevTimers[i] );
//				}
//
//			}
//
//			deferred.resolve();
//
//			return pipe;
//
//		}
//
//	};

	/**
	 * if clipboard isnt empty, display confirmation dialogue
	 *
	 * @returns {jQueryDeferred}
	 */
	this.checkClipboard = function() {

		var deferred = $.Deferred();
		var i, len;

		if ( this.clipboard.channels.length ) {
			new jkqDialog({
				title:			this.i18.dialog.del_clipboard.title,
				text:			this.i18.dialog.del_clipboard.text.format( this.clipboard.channels.length ),
				autoSkipText:	tma.i18.respond_auto_skip,
				autoSkipButton:	'clsDeleteClipboard',
				buttons:		{
					jkqDialogCancel:	tma.i18.respond_cancel,
					clsDeleteClipboard:	this.i18.button.del.text
				}
			}).confirm(function(button) {

				deferred.resolve();

			}).cancel(function() {

				deferred.reject();

			});

		} else {

			// nothing to do
			deferred.resolve();

		}

		return deferred.promise();

	};

//	var osdTimer;
//
//	self.syncSEchannels = function() {
//
//		// send menu key to open SmartEPG settings
//		tma.sendremote( 'menu' );
//
//		// after 1 second open SE channel list by 'OK' button
//		window.setTimeout(function() {
//			tma.sendremote( 'ok' );
//		}, 1000);
//
//		$('.cls-pvrosd').attr( 'src', tma.PVRosd + "#e" + Date.now() );
//		osdTimer = window.setInterval(function() {
//			$('.cls-pvrosd').attr( 'src', tma.PVRosd + "#e" + Date.now() );
//		},2000);
//
//	};

	/**
	 * execute the channel update chain
	 *
	 * @param {string}	filename	desired file name without path
	 * @param {TmaTap}	se
	 */
	this.updatePvrChannels = function( filename, se ) {

//		// 1) refresh timer list
//		updateStatus.refreshTimers = "exec";
//		tma.timers.load().done(function() {
//			updateStatus.refreshTimers = "ok";
//			updateStatus.prevTimersCnt = tma.timers.timers.length;

			// 2) stop SmartEPG if running
			updateStatus.stopSE = "exec";
			self.stopSE( se ).done(function() {
				updateStatus.stopSE = "ok";

				// 3) start ChannelListSaver to import channels
				updateStatus.startCLS = "exec";
				self.startCLS( filename, CLS_IMPORT, CLS_FMT_TXT, CLS_SILENT ).done(function() {

					updateStatus.startCLS = "ok";

					// cls deleted all timers => remove timers from list
					// !!! cls now handles timers
//					var timers = tma.timers.timers.splice(0, tma.timers.timers.length);

					// update channel numbers
					for( var i = 0, len = self.channels.channels.length; i < len; i++) {
						var clsChannel = self.channels.channels[i];
						if (
								(clsChannel.channel.chtype == TmaChannel.TYPE_TV) ||
								(clsChannel.channel.chtype == TmaChannel.TYPE_RADIO)
						) {
							// not an analogue channel
							clsChannel.channel.chnum = clsChannel.number - 1;
						}
					}

					// hide window close button, because app has to be restarted
					$('.cls-update').parents('.jkqWindow').find('.jkqWindowClose').hide();

					// 4.) start SmartEPG, if it was previously running
					updateStatus.startSE = "exec";
					self.startSE( se ).done(function() {
						updateStatus.startSE = "ok";

						// !!! cls now handles timers
//						updateStatus.createTimers = "ok";
						$('.cls-cancel-update').prop('disabled', false);
//						if ( !se ) {
//							// remove timers of deleted channels
//							timers = timers.filter( function( timer ) {
//								var channel = self.channels.channel( timer.logoid );
//								return	(channel instanceof ClsChannel);
//							});
//						}

//						// 5) if SE was previously running, wait for restored timers, else recreate timers
//						updateStatus.createTimers = "exec";
//						self.waitForOrRecreateTimers( se, timers ).done(function() {
//
//							updateStatus.createTimers = "ok";
//
////							if (se) {
////								// 6. synchronise SmartEPG channel list by remote keys
////								updateStatus.syncSE = "exec";
////								//self.syncSEchannels();
////								// User has to sync SE channels manually
////								osdTimer = window.setInterval(function() {
////									$('.cls-pvrosd').attr( 'src', tma.PVRosd + "#e" + Date.now() );
////								}, 1000);
////							}
//
//							$('.cls-cancel-update').prop('disabled', false);
//
//						}).fail(function() {
//							// waiting for restored/recreated timers failed
//							updateStatus.createTimers = "fail";
//							$('.cls-cancel-update').prop('disabled', false);
//						});

					}).fail(function() {
						// Starting SE failed
						updateStatus.startSE = "fail";
						$('.cls-cancel-update').prop('disabled', false);
					});

				}).fail(function( reason ) {

					// self.startCLS failed

					updateStatus.startCLS = "fail";

					if ( reason == CLS_ERROR ) {

						// ChannelListSaver.tap did not return with result=success

						// hide window close button, because app has to be restarted
						$('.cls-update').parents('.jkqWindow').find('.jkqWindowClose').hide();

						// restore previous state
						self.startCLS( 'Settings_vor.std', CLS_IMPORT, CLS_FMT_SYS, CLS_SILENT ).always(function() {

							// let's start SmartEPG again
							updateStatus.startSE = "exec";
							self.startSE( se ).done(function() {
								updateStatus.startSE = "ok";
							}).fail(function() {
								updateStatus.startSE = "fail";
							}).always(function() {

								// !!! cls now handles timers
								// if SE was previously running, wait for restored timers, else recreate timers
//								updateStatus.createTimers = "ok";
								$('.cls-cancel-update').prop('disabled', false);
//								updateStatus.createTimers = "exec";
//								self.waitForOrRecreateTimers( se, timers ).done(function() {
//									updateStatus.createTimers = "ok";
//								}).fail(function() {
//									updateStatus.createTimers = "fail";
//								}).always( function() {
//									$('.cls-cancel-update').prop('disabled', false);
//								});

							});

						});

					} else if (
							(reason == CLS_FAILED) ||
							(reason == CLS_TIMEOUT)
					) {

						// starting ChannelListSaver.tap failed
						// or
						// finishing ChannelListSaver.tap failed (should not occur)

						// let's start SmartEPG again
						updateStatus.startSE = "exec";
						self.startSE( se ).done(function() {
							updateStatus.startSE = "ok";
						}).always( function() {
							$('.cls-cancel-update').prop('disabled', false);
						});

					}

				});

			}).fail(function() {
				// stopping SmartEPG or refreshing timers failed
				updateStatus.stopSE = "fail";
				$('.cls-cancel-update').prop('disabled', false);
			});

//		}).fail(function() {
//			// refreshing timer list failed
//			updateStatus.refreshTimers = "fail";
//			$('.cls-cancel-update').prop('disabled', false);
//		});

	};

	/**
	 * based on
	 * https://github.com/eligrey/FileSaver.js
	 */
	this.exportChannels = function() {

		var content = this.createClsChannelsFile();
		var mimeType = 'text/plain';
		var fileName = 'Channels.txt';

        // saving text file
        var blob = new Blob([content], {type: "text/plain"});
        saveAs(blob, fileName);

	};

	/**
	 * read channel list from local CLS channels file
	 *
	 * @param {File}	file
	 * @returns {jQueryDeferred}
	 */
	this.importChannels = function( file ) {

		var deferred = $.Deferred();

		this.resetChannels();

		var fileReader = new FileReader();
		fileReader.onload = function() {

		    var text = new Uint8Array( this.result );
			deferred.resolve( text );

		};
		fileReader.readAsArrayBuffer( file );

		return deferred.promise();

	};

	/**
	 * upload text file and display update dialogue
	 *
 	 * @param {Uint8Array} text
	 */
	this.updateDialogue = function( text ) {

		// test, if SmartEPG if running
		var se = tma.taps.find( 0x8E0A424E );
		updateStatus.seWasRunning = !!se;

		var popup = new jkqWindow();
		popup.width = 550;

		popup.showModal();
		popup.setTitle( this.i18.update.title );

		var $template = JK.template(
				this.templates.update,
				{
					i18:		this.i18,
					i18g:		tma.i18,
					status:		updateStatus,
					osd_url:	tma.PVRosd
				}
		).appendTo(popup.container.find('DIV.jkqWindowContent'))

		.on('click', '.cls-start-update', function() {
			$(this).prop('disabled', true);
			$('.cls-cancel-update')
			.prop('disabled', true)
			.text(self.i18.button.close.text);
			var filename = $(this).prop('filename');
			self.updatePvrChannels( filename, se );
		})

		.on('click', '.cls-cancel-update', function() {
			if (updateStatus.startCLS != "none") {
				// restart app with default page
				window.location.hash = '';
				window.location.reload();
			} else {
				popup.close();
			}
		})

//		.on('click', '.cls-remote-menu', function() {
//			tma.sendremote( 'menu' );
//		})
//
//		.on('click', '.cls-remote-ok', function() {
//			tma.sendremote( 'ok' );
//		})
//
//		.on('click', '.cls-remote-exit', function() {
//			tma.sendremote( 'exit' );
//		})

		;

		// upload new channels file
		updateStatus.saveTextfile = "exec";
		var filename = 'channels_' + tma.date('Y-m-d_H-i-s') + '.webc.txt';
		self.saveChannels( filename, text ).done(function() {
			updateStatus.saveTextfile = "ok";
			$('.cls-start-update').prop('disabled', false);
			$('.cls-start-update').prop('filename', filename);
		}).fail(function() {
			updateStatus.saveTextfile = "fail";
		});

		popup.on( 'close', function() {
			JK.removeTemplate( $template, true );
		} );

		popup.adjustPos();

	};

}
